/*
 * decls.h
 *
 * Declarations of stuff needed to drive the digital logic simulator.
 * [eichin:19880311.1902EST]
 */

#define ASSERT(x)
#define ROPE_SIZE 16
#include <msimmacs.h>

typedef struct _rope_elem rope_elem, *ropeptr;


typedef ropeptr Timeline;	/* tags = Event_val, data = time */
typedef ropeptr Horizon;	/* points to end of Timeline... */
typedef ropeptr Generics;	/* tags = FN | WIRE, data = *fn(), *wire */
typedef ropeptr Deps;		/* tags = deps, data = *Depend */
typedef ropeptr Event;
typedef ropeptr Wire;		/* tags = DEPS | CUR | PAST | CONT,
				 * data = *Deps, *Timeline, *Contend
				 */
typedef ropeptr Contention;	/* tags = Event_tag, vals = *wire */
typedef ropeptr Wirelist;	/* tags = Wire, data = wire */
/*
 * special tags for allocator structures.
 */

/*
 * Here, END==0 would be a good efficiency hack, but we couldn't keep
 * it isolated from the other types; thus, we make it 1024 for now. In
 * the future, we should probably just start the others at a special
 * high value...
 * [eichin:19880330.1624EST]
 */

typedef enum _ctl_tag {
  END=1024, BEGIN, NEXT_FORWARD, NEXT_BACKWARD,
  NEXT_FORWARD_UNALIGNED, NEXT_BACKWARD_UNALIGNED,
  FREELIST, __notctl,
} Ctl_tag;



typedef enum _Ctl_data {
  end=__notctl, notend,
  __notctl_data,
} Ctl_data;

typedef ropeptr Next_data;

/*
 * WFile
 * File structure with buffering needed for fast input from standard
 * input stream devices.
 */
#include <stdio.h>

typedef long Time;
extern Time globaltime;

typedef struct _WFile {
  FILE	*filedesc;
  Time	next;
} WFile, *WFileptr;

/*
 * Event
 * The smallest piece of the simulation, the Event is a combination of
 * an assertion (not necessarily a change?) of state and a time stamp
 * indicating when this occurs.
 */
typedef enum _Event_tag {
  tag_init=__notctl_data,
  LOGIC_HI, LOGIC_LOW, 
  EARLY_HI, EARLY_LOW,
  TRISTATE, CONTENDING,
  UNDEFINED, FILE_FETCH,
  __notevent,
} Event_tag;

typedef Time Event_data;

/*
 * Contention
 * A contention record holds all of the states and all of the wires
 * that assert them, so that they can be deleted and added as time
 * passes. A contention procedure can be defined somewhere (per wire?
 * per system? or for more realism, per input circuit?) for deriving
 * Current from Contention.
 */

typedef Event_tag Contention_tag;
typedef Wire *Contention_data;	/* ???[eichin:19880401.2023EST] */

/*
 * Depend
 * A Depend is a circuit element that takes a given wire as input. It
 * carries with it the list of wires it outputs to, and the procedure
 * to actually process the wire input.
 */
typedef int (*PFI)();

/*
 * Deps
 * A Deps (Dependencies) element is simply a collection of all the
 * Depends which are dependent upon this wire.
 */
typedef enum _Deps_tag {
  DEPSLIST=__notevent,
  __notdeps,
} Deps_tag;

typedef union _Deps_data {
  Generics	generics;
} Deps_data;

/*
 * Wire
 * A Wire is a model for an actual circuit path. It consists of the
 * Deps (all circuit elements hooked to the wire), the Future (all
 * events scheduled to come) the Past (disposable logs of where the
 * circuit has been) the Now (an easy to find current value of the
 * wire, it should be an Event) and the Contention (a full description
 * of the current state of the wire, for Conflict resolution.)
 */
typedef enum _Wire_tag {
  SYMNAME=__notdeps,
  DEPENDS,
  NOW, THEN, FUTURE,
  PAST, CONTENTION,
  __notwire,
} Wire_tag;

typedef union _Wire_data {
  char		*symname;
  Deps		deps;
  Timeline	now;		/* what is evalled */
  Timeline	then;		/* what is scanned */
  Timeline	future;		/* what the above point into. */
  Timeline	past;
  Contention	contention;
} Wire_data;

/*
 * Wirelist
 * A Wirelist is the master list tying together all the wires in a
 * circuit. This drives the simulation. A special element (WFile)
 * should be inserted once (per input stream) and contains pointers to
 * the input file and info for the conversion of it into actual Events.
 */
typedef enum _Wirelist_tag {
  WIRE=__notwire, WFILE,
  __notwirelist,
} Wirelist_tag;

typedef union _Wirelist_data {
  Wire		wire;
  WFileptr	wfile;
} Wirelist_data;

/*
 * Another type, just to keep the argument lists interesting.
 * Specifically, `integer argument' for arbitrary things in dependency
 * lists. While I'm at it, call this `generic' and give it lots of
 * pieces... [eichin:19880404.0632EST]
 */

typedef enum _Generic_tag {
  GENINT=__notwirelist,
  GENSTRING,
  GENTIME,
  GENWIRE,
  GENFUNC,
  __notgeneric,
} Generic_tag;

typedef union _Generic_data {
  int		genint;
  char *	genstring;
  Time		gentime;
  Wire		genwire;
  PFI		genfunc;
} Generic_data;
    

/*
 * These structures are set up to keep the rope structure valid.
 */ 

typedef union _rope_tag {
  Ctl_tag	CTL;
  Event_tag	EVENT;
  Deps_tag	DEPS;
  Wire_tag	WIRE;
  Wirelist_tag	WIRELIST;
  Generic_tag	GENERIC;
} rope_tag;

typedef union _rope_data {
  Ctl_data	CTL;
  Next_data	NEXT;
  Event_data	EVENT;
  Deps_data	DEPS;
  Wire_data	WIRE;
  Wirelist_data	WIRELIST;
  Generic_data	GENERIC;
} rope_data;

struct _rope_elem {
  rope_tag	tag;
  rope_data	data;
};

#include <msimexts.h>

ropeptr getblock();
Wirelist make_wirelist_1();
Wirelist make_wirelist_2();
Wirelist make_wirelist_3();
Wire makewire();
ropeptr brk_and_format_block();
extern int debug;
